import asyncio

import polars

import cameo_claw as cc
from cameo_claw import mkdir
from cameo_claw.client import Client
from cameo_claw.net import url_to_filename
from cameo_claw.worker_f import pipe_download_iot_csv_gz
import os

lst_device_id = [
    '10180507985',
    '10183804818',
    '10180935138',
    '10182362729',
    '10186915768',
    '10182765572',
    '10183643733',
    '10187817545',
    '10180711468',
    '10181888891',
    '10184237251',
    '10200376949',
    '10187714974',
    '10188959173',
    '10181133624',
    '10189875513',
    '10184780057',
    '10187536226',
    '10188542674',
    '10187616860',
    '10190212306',
    '10200485379',
    '10184404771',
    '10188107333',
    '10186524358',
    '10187942042',
    '10183163884',
    '10180654119',
    '10185092420',
    '10187452115',
    '10190000864',
    '10190479538',
    '10185579073',
    '10181983795',
    '10188844366',
    '10189360662',
    '10182995926',
    '10181055695',
    '10189089883',
    '10181464107',
    '10184979038',
    '10181258942',
    '10184871059',
    '10185137651',
    '10182074994',
    '10190374312',
    '10186752944',
    '10189461434',
    '10184589018',
    '10181395211',
    '10186372622',
    '10186223376',
    '10182118699',
    '10190513901',
    '10185716972',
    '10189779191',
    '10188219952',
    '10185390063',
    '10181621345',
    '10182567005',
    '10186123517',
    '10188652949',
    '10184615499',
    '10189289540',
    '10182827296',
    '10185414614',
    '10184032107',
    '10185295344',
    '10182603505',
    '10186080368',
    '10187340888',
    '10186497242',
    '10187268812',
    '10183739942',
    '10188795449',
    '10183269165',
    '10183435474',
    '10180884769',
    '10188098571',
    '10183378041',
    '10185861505',
    '10183938343',
    '10185920135',
    '10181700710',
    '10188450871',
    '10186643177',
    '10184128532',
    '10190122978',
    '10188364226',
    '10184373796',
    '10186871514',
    '10181510720',
    '10183091769',
    '10187195615',
    '10182493773',
    '10187012610',
    '10182273531',
    '10183554350',
    '10185613026',
    '10189170767',
]


def get_lst_path_taipei_100(directory='./data/topic_pipe_download/', ext='.csv.gz'):
    lst_path = []
    for device_id in lst_device_id:
        for i in range(1, 32):  # range(1, 32)
            day = f'2022-03-{i:02}'
            path = f'{directory}device_{device_id}_daily_{day}{ext}'
            lst_path.append(path)
    return lst_path


def get_lst_url_taipei_100():
    return get_lst_url_by_date(lst_device_id)


def get_lst_url_by_date(lst_device_id):
    lst_url = []
    for device_id in lst_device_id:
        for i in range(1, 32):  # range(1, 32)
            day = f'2022-03-{i:02}'
            url = f'https://iot.epa.gov.tw/fapi_open/topic-device-daily/topic_save.industry.rawdata.material/device_{device_id}/device_{device_id}_daily_{day}.csv.gz'
            lst_url.append(url)
    return lst_url


def pipe_download_iot_zip(url, d):
    try:
        df = cc.read_csv(url)
        df = cc.distinct(df, d['lst_distinct_column'])
        df = cc.filter(df, d['lst_column_match'])
        df = cc.sort(df, d['sort_column'])
        g = cc.groupby(df, d['lst_group_by_column'])
        for df in g:
            cc.append_csv(df, d)
        return {'is_success': True, 'url': url}
    except Exception as e:
        return {'is_success': False, 'url': url, 'exception': e}


def test_pipe_download_iot_zip():
    d = {'lst_url': get_lst_url_taipei_100(),
         'target_directory': './data/topic_download_iot_zip/',
         'lst_distinct_column': ['deviceId', 'localTime', 'sensorId'],
         'lst_column_match': [['sensorId', 'pm2_5'], ['sensorId', 'voc']],
         'sort_column': 'localTime',
         'lst_group_by_column': ['deviceId', 'sensorId'],
         'append_column': 'deviceId',
         'column_header': 'createTime,deviceId,lat,localTime,lon,sensorId,value\n'}
    if cc.has_zip_cache(d):
        zip_cache_path = cc.zip_cache_path(d)
        print('zip_cache_path existed:', zip_cache_path)
        return zip_cache_path
    for int_progress, dic_result in cc.pipe(pipe_download_iot_zip, d['lst_url'], d):
        if not dic_result['is_success']:
            print(f"test_epa_download, e: {dic_result['exception']}, url: {dic_result['url']}'")
    return cc.zip_cache(d)


def test_pipe_download_iot_csv_gz():
    d = {'lst_url': get_lst_url_taipei_100(),
         'target_directory': './data/topic_download_iot_csv_gz/',
         'lst_distinct_column': ['deviceId', 'localTime', 'sensorId'],
         'lst_column_match': [['sensorId', 'pm2_5'], ['sensorId', 'voc']],
         'sort_column': 'localTime',
         'lst_group_by_column': ['deviceId', 'sensorId']}
    for int_progress, dic_result in cc.pipe(pipe_download_iot_csv_gz, d['lst_url'], d):
        if not dic_result['is_success']:
            print(f"test_epa_download, e: {dic_result['exception']}, url: {dic_result['url']}'")


# def pipe_nation(url, d):
#     try:
#         df = cc.read_csv(url)
#         df = cc.distinct(df, d['lst_distinct_column'])
#         df = cc.filter(df, d['lst_column_match'])
#         df = cc.sort(df, d['sort_column'])
#         g = cc.groupby(df, d['lst_group_by_column'])
#         for df in g:
#             cc.to_csv_gz(df, d['target_directory'], url, d['lst_group_by_column'])
#         return {'is_success': True, 'url': url}
#     except Exception as e:
#         return {'is_success': False, 'url': url, 'exception': e}
#
#
# def test_nation():
#     d = {'lst_url': [
#         'https://iot.epa.gov.tw/fapi_open/topic-device-daily/topic_save.nation.rawdata.material/device_11647957860/device_11647957860_daily_2022-04-21.csv.gz'],
#         'target_directory': './data/topic_nation/',
#         'lst_distinct_column': ['deviceId', 'localTime', 'sensorId'],
#         'lst_column_match': [['sensorId', 'pm2_5'], ['sensorId', 'voc']],
#         'sort_column': 'localTime',
#         'lst_group_by_column': ['deviceId', 'sensorId']}
#     for int_progress, dic_result in cc.pipe(pipe_nation, d['lst_url'], d):
#         if not dic_result['is_success']:
#             print(f"test_nation, e: {dic_result['exception']}, url: {dic_result['url']}'")


def pipe_nation_zip(url, d):
    try:
        df = cc.read_csv(url)
        df = cc.distinct(df, d['lst_distinct_column'])
        df = cc.filter(df, d['lst_column_match'])
        df = cc.sort(df, d['sort_column'])
        g = cc.groupby(df, d['lst_group_by_column'])
        for df in g:
            cc.append_csv(df, d)
        cc.append_csv(df, d)
        return {'is_success': True, 'url': url}
    except Exception as e:
        return {'is_success': False, 'url': url, 'exception': e}


def test_nation_zip():
    d = {'lst_url': [
        'https://iot.epa.gov.tw/fapi_open/topic-device-daily/topic_save.nation.rawdata.material/device_10841800054/device_10841800054_daily_2022-04-22.csv.gz'],
        'target_directory': './data/topic_nation_zip/',
        'lst_distinct_column': ['deviceId', 'localTime', 'sensorId'],
        'lst_column_match': [['sensorId', 'pm2_5'], ['sensorId', 'voc']],
        'sort_column': 'localTime',
        'lst_group_by_column': ['deviceId', 'sensorId'],
        'append_column': 'deviceId',
        'column_header': 'attributes,createTime,deviceId,lat,localTime,lon,sensorId,value\n'}
    if cc.has_zip_cache(d):
        zip_cache_path = cc.zip_cache_path(d)
        print('zip_cache_path existed:', zip_cache_path)
        return zip_cache_path
    for int_progress, dic_result in cc.pipe(pipe_nation_zip, d['lst_url'], d):
        if not dic_result['is_success']:
            print(f"test_epa_download, e: {dic_result['exception']}, url: {dic_result['url']}'")
    return cc.zip_cache(d)


async def test_client(api_url):
    client = Client(api_url)
    d = {
        'target_directory': './data/topic_download_iot_csv_gz/',
        'lst_distinct_column': ['deviceId', 'localTime', 'sensorId'],
        'lst_column_match': [['sensorId', 'pm2_5'], ['sensorId', 'voc']],
        'sort_column': 'localTime',
        'lst_group_by_column': ['deviceId', 'sensorId']}
    lst_result = await client.map(pipe_download_iot_csv_gz, get_lst_url_taipei_100(), d, int_chunk_size=100)


def pipe_download(url, d):
    try:
        cc.a_download(url, d['target_directory'])
        return {'is_success': True, 'url': url}
    except Exception as e:
        return {'is_success': False, 'url': url, 'exception': e}


def test_pipe_download():
    d = {'lst_url': get_lst_url_taipei_100(),
         'target_directory': './data/topic_pipe_download/',
         }
    for int_progress, dic_result in cc.pipe(pipe_download, d['lst_url'], d):
        if not dic_result['is_success']:
            print(f"test_epa_download, e: {dic_result['exception']}, url: {dic_result['url']}'")


def pipe_local(path, d):
    try:
        df = polars.read_csv(path)
        df = cc.distinct(df, d['lst_distinct_column'])
        df = cc.filter(df, d['lst_column_match'])
        df = cc.sort(df, d['sort_column'])
        g = cc.groupby(df, d['lst_group_by_column'])
        for df in g:
            cc.to_csv_gz(df, d['target_directory'], path, d['lst_group_by_column'])
        return {'is_success': True, 'url': path}
    except Exception as e:
        return {'is_success': False, 'url': path, 'exception': e}


def test_local():
    d = {'target_directory': './data/topic_download_iot_csv_gz/',
         'lst_distinct_column': ['deviceId', 'localTime', 'sensorId'],
         'lst_column_match': [['sensorId', 'pm2_5'], ['sensorId', 'voc']],
         'sort_column': 'localTime',
         'lst_group_by_column': ['deviceId', 'sensorId']}
    for int_progress, dic_result in cc.pipe(pipe_local, get_lst_path_taipei_100(), d):
        if not dic_result['is_success']:
            print(f"test_epa_download, e: {dic_result['exception']}, url: {dic_result['url']}'")


def test_local_csv():
    d = {'target_directory': './data/topic_download_iot_csv_gz/',
         'lst_distinct_column': ['deviceId', 'localTime', 'sensorId'],
         'lst_column_match': [['sensorId', 'pm2_5'], ['sensorId', 'voc']],
         'sort_column': 'localTime',
         'lst_group_by_column': ['deviceId', 'sensorId']}
    for int_progress, dic_result in cc.pipe(pipe_local, get_lst_path_taipei_100('./data/topic_csv/', '.csv'), d):
        if not dic_result['is_success']:
            print(f"test_epa_download, e: {dic_result['exception']}, url: {dic_result['url']}'")


def pipe_local_csv_to_parquet(path, d):
    try:
        df = polars.read_csv(path)
        to_path = d['target_directory'] + url_to_filename(path, is_ext=False) + '.parquet'
        df.write_parquet(to_path, compression="snappy")
        return {'is_success': True, 'path': path}
    except Exception as e:
        return {'is_success': False, 'path': path, 'exception': e}


def test_local_csv_to_parquet():
    d = {'target_directory': './data/topic_parquet/'}
    mkdir('./data/topic_parquet/')
    for int_progress, dic_result in cc.pipe(pipe_local_csv_to_parquet,
                                            get_lst_path_taipei_100('./data/topic_csv/', '.csv'), d):
        if not dic_result['is_success']:
            print(f"test_epa_download, e: {dic_result['exception']}, path: {dic_result['path']}'")


def pipe_local_parquet_compute(path, d):
    try:
        df = polars.read_parquet(path)
        df = cc.distinct(df, d['lst_distinct_column'])
        df = cc.filter(df, d['lst_column_match'])
        df = cc.sort(df, d['sort_column'])
        g = cc.groupby(df, d['lst_group_by_column'])
        for df in g:
            cc.to_parquet(df, d['target_directory'], path, d['lst_group_by_column'])
            # cc.to_csv_gz(df, d['target_directory'], path, d['lst_group_by_column'])
        return {'is_success': True, 'path': path}
    except Exception as e:
        return {'is_success': False, 'path': path, 'exception': e}


def test_local_parquet_compute():
    d = {'target_directory': './data/topic_parquet_compute/',
         'lst_distinct_column': ['deviceId', 'localTime', 'sensorId'],
         'lst_column_match': [['sensorId', 'pm2_5'], ['sensorId', 'voc']],
         'sort_column': 'localTime',
         'lst_group_by_column': ['deviceId', 'sensorId']}
    mkdir('./data/topic_parquet_compute/')
    for int_progress, dic_result in cc.pipe(pipe_local_parquet_compute,
                                            get_lst_path_taipei_100('./data/topic_parquet/', '.parquet'), d):
        if not dic_result['is_success']:
            print(f"test_epa_download, e: {dic_result['exception']}, path: {dic_result['path']}'")


if __name__ == '__main__':
    # test_pipe_download_iot_zip()
    # test_pipe_download_iot_csv_gz()
    # test_nation_zip()
    # test_pipe_download() # 純下載無計算
    # asyncio.run(test_client('https://sc8.bowenchiu.repl.co/api/'))
    # test_local()
    # test_local_csv()
    # test_local_csv_to_parquet()
    test_local_parquet_compute()



import sys

from cameo_claw import lst_demo_url

help_message = f'''
python3 -m pip install cameo_claw --upgrade
python3 -c "
from cameo_claw import it_filter_group
lst_url={lst_demo_url}
target_directory = './data/topic_filter_group/'
int_total=len(lst_url)
lst_distinct_column = ['deviceId', 'localTime', 'sensorId']
lst_column_match = [['sensorId', 'pm2_5'], ['sensorId', 'voc']]
sort_column = 'localTime'
lst_group_by_column = ['deviceId']
for int_progress, done_url in it_filter_group(lst_url, target_directory, lst_distinct_column, lst_column_match, sort_column, lst_group_by_column):
        pass
"
'''


def show_help():
    if len(sys.argv) == 1:
        print(help_message)


if __name__ == '__main__':
    show_help()



import cameo_claw as cc


def pipe_download_iot_csv_gz(url, d):
    try:
        # print('debug worker_f.py pipe_download_iot_csv_gz d', d)
        df = cc.read_csv(url)
        df = cc.distinct(df, d['lst_distinct_column'])
        df = cc.filter(df, d['lst_column_match'])
        df = cc.sort(df, d['sort_column'])
        g = cc.groupby(df, d['lst_group_by_column'])
        for df in g:
            cc.to_csv_gz(df, d['target_directory'], url, d['lst_group_by_column'])
        return {'is_success': True, 'url': url}
    except Exception as e:
        return {'is_success': False, 'url': url, 'exception': e}
        #@@ todo debug 問題抓到了就在此，因為 e 無法 http post 回傳給 scheduler



import gzip
from io import BytesIO
import polars as pl
from fastapi import FastAPI
from starlette.middleware.cors import CORSMiddleware

from cameo_claw.file import mkdir
from cameo_claw.net import url_to_filename, requests_get_ram_cache

import warnings
# import pandas as pd
from cameo_claw.functional import it_mp_f

warnings.filterwarnings("ignore")


def bytesio_to_gzip(df, url, target_directory):
    bytesio, filename = bytesio_filename(df, url)
    path = f'{target_directory}{filename}.csv.gz'
    write_gzip(bytesio, path)
    return url


def requests_get_write(f_write, target_directory, url):
    try:
        mkdir(target_directory)
        return requests_get_ram_cache(f_write, url, target_directory)
    except Exception as e:
        print(f'cameo_claw.py,requests_get_write,Exception:{e, url}')


def bytesio_filename(df, url, has_header=True):
    bytesio = BytesIO()
    df_write_csv(df, bytesio, has_header=has_header)
    filename = url_to_filename(url)
    return bytesio, filename


def df_write_csv(df, bytesio, has_header):
    '''
    fix polars bug
    polars has a bug, that iot attributes first byte lack a double quote " like this:
    attributes,createTime,deviceId,lat,localTime,lon,sensorId,value
    [{""key"":""ValueRaw"",""value"":""88.7""},{""key"":""Status"",""value"":""0""},{""key"":""ValueSD"",""value"":""0""},{""key"":""CollectRate"",""value"":""100.0""},{""key"":""Valid"",""value"":""0""}]",2022-04-21 00:02:00.042,11647957860,23.8549,2022-04-21 00:01:00.000,120.2696,rh,88.7
    "[{""key"":""ValueRaw"",""value"":""27""},{""key"":""Status"",""value"":""64""},{""key"":""ValueSD"",""value"":""0""},{""key"":""CollectRate"",""value"":""0.0""},{""key"":""Valid"",""value"":""4""}]",2022-04-21 00:02:00.042,11647957860,23.8549,2022-04-21 00:01:00.000,120.2696,pm2_5,27.0
    '''
    if df.row(0)[0][0] == '[':
        bytesio.write(b'"')
    return df.write_csv(bytesio, has_header=has_header)


def write_gzip(bytesio, path, mode='wb'):
    with gzip.open(path, mode) as f:
        f.write(bytesio.getvalue())
    return True


def pl_read_csv_distinct(bytes1, lst_distinct_column):
    df = pl.read_csv(bytes1, infer_schema_length=20000)
    df = df.distinct(subset=lst_distinct_column)
    return df


def pl_read_csv_str(bytes1):
    return pl.read_csv(bytes1, infer_schema_length=0).with_columns(pl.all().cast(pl.Int32, strict=False))


# def pd_read_csv(bytes1):
#     df = None
#     if bytes1[0:2] == b'\x1f\x8b':
#         df = pd.read_csv(BytesIO(bytes1), compression='gzip')
#     else:
#         df = pd.read_csv(BytesIO(bytes1))
#     return df


def groupby_write_gzip(df, url, lst_group_by_column, target_directory, has_header=True):
    for df in df.groupby(lst_group_by_column):
        df_write_gzip(df, has_header, lst_group_by_column, target_directory, url)
    return url


def df_write_gzip(df, has_header, lst_group_by_column, target_directory, url):
    filename_tail = '_group_' + '-'.join(list(
        map(lambda column:
            str(df.row(0)[df.find_idx_by_name(column)]).replace('_', '-'),
            lst_group_by_column)))
    bytesio, filename = bytesio_filename(df, url, has_header)
    path = f'{target_directory}{filename}{filename_tail}.csv.gz'
    write_gzip(bytesio, path)
    return filename_tail


def it_f(f, lst_url, target_directory, lst_distinct_column, lst_column_match, sort_column):
    return it_mp_f(f, [tuple([url, target_directory, lst_distinct_column,
                              lst_column_match, sort_column]) for url in lst_url])


def condition_filter_sort(df, lst_column_match, sort_column):
    df = df.filter(eval(condition(lst_column_match)))
    if sort_column:
        df = df.sort(sort_column)
    return df


def condition(lst_column_match):
    c = ''
    for lst_col_val in lst_column_match:
        c += f"""(pl.col('{lst_col_val[0]}')=='{lst_col_val[1]}') | """
    c = c[:-3]
    return c


def distinct_filter_sort(bytes1, lst_distinct_column, lst_column_match, sort_column):
    df = pl_read_csv_distinct(bytes1, lst_distinct_column)
    df = condition_filter_sort(df, lst_column_match, sort_column)
    return df


def fastapi_app():
    app = FastAPI(docs_url="/d", redoc_url=None, openapi_url="/o.json")
    app.add_middleware(CORSMiddleware,
                       allow_origins=['*'],
                       allow_credentials=True,
                       allow_methods=['*'],
                       allow_headers=['*'])
    return app


lst_demo_url = [
    'https://cameo-claw-data.vercel.app/device_21152332479_daily_2022-03-30.csv',
    'https://cameo-claw-data.vercel.app/device_21152332479_daily_2022-03-31.csv']


# 2022-04-21 bowen, copy from https://appdividend.com/2021/06/15/how-to-split-list-in-python/
def split_list(lst, n):
    yield from [lst[i: n + i] for i in range(0, len(lst), n)]
